"""
Top level module for dynaddrmgr application.

Classes:
    TemplateManager
"""
import filecmp
from os import R_OK, W_OK, access
from pathlib import Path
from shutil import copy2
from tempfile import NamedTemporaryFile
from typing import Optional

from jinja2 import Template

from wtforglib.commander import Commander
from wtforglib.files import verify_directory
from wtforglib.fstats import set_owner_group_perms
from wtforglib.kinds import StrAnyDict
from wtforglib.options import OptionsDict
from wtforglib.scribe import Scribe
from wtforglib.versioned import unlink_path
from wtforglib.versionfile import version_file

KDEST = "dest"


class TemplateWriter(Commander):  # noqa: WPS214
    """
    Template generator.

    This class is responsible for managing the templates specified
    in the configuration file.
    """

    changed: bool

    def __init__(
        self,
        opts: Optional[OptionsDict] = None,
        scribe: Optional[Scribe] = None,
    ) -> None:
        """Constructs a new instance of TemplateWriter.

        Parameters
        ----------
        opts : Optional[OptionsDict], optional
            Options dictionary, by default None
        scribe : Optional[Scribe], optional
            Scribe for logging, by default None
        """
        super().__init__(opts)
        self.changed = False
        if scribe is not None:
            self.scribe = scribe

    def generate(
        self,
        tmpl_name: str,
        tmpl_value: StrAnyDict,
        tmpl_var: StrAnyDict,
    ) -> int:
        """Generates the template if needed.

        Parameters
        ----------
        tmpl_name : str
            Template name
        tmpl_value : StrAnyDict
            Template information
        tmpl_var : StrAnyDict
            Template variable data

        Returns
        -------
        int
            Exit code
        """
        self.changed = False
        if self._verify_config_data(tmpl_name, tmpl_value):
            return self._update_template(tmpl_value, tmpl_var)
        return 1

    def _update_template(self, tmpl_value: StrAnyDict, tmpl_var: StrAnyDict) -> int:
        """Update the template if necessary.

        Parameters
        ----------
        tmpl_value : StrAnyDict
            Data describing the target file requirements
        tmpl_var : StrAnyDict
            Variables used by the template

        Returns
        -------
        int
            exit status
        """
        dest = tmpl_value.get(KDEST, "")
        self.changed = self._render_template(
            tmpl_value,
            tmpl_var,
        )
        if self.changed:
            self.info("Updated: {0}".format(dest))
            if not self.istest():
                set_owner_group_perms(
                    dest,
                    tmpl_value.get("owner", ""),
                    tmpl_value.get("group", ""),
                    tmpl_value.get("mode", ""),
                )
                cargs = tmpl_value.get("on_changed", [])
                if cargs:
                    cmdres = self.run_command(tuple(cargs))
                    return cmdres.returncode
        else:
            self.debug("Unchanged dest: {0}".format(dest))
        return 0

    def _verify_config_data(self, tmpl_name: str, tmpl_value: StrAnyDict) -> bool:
        """Verifies the configuration.

        Parameters
        ----------
        tmpl_name : str
            Name of the template
        tmpl_value : StrAnyDict
            Template values

        Returns
        -------
        bool
            True if the configuration is valid
        """
        if self._verify_template_required_keys(tmpl_name, tmpl_value):
            if self._verify_template_source(tmpl_value.get("src", "")):
                if self._verify_target(tmpl_value.get(KDEST, "")):
                    return True
        return False

    def _render_template(
        self,
        tmpl_value: StrAnyDict,
        tmpl_var: StrAnyDict,
    ) -> bool:
        """Render the template to the file system.

        Parameters
        ----------
        tmpl_value : StrAnyDict
            Data describing the target file requirements
        tmpl_var : StrAnyDict
            Variables used by the template

        Returns
        -------
        int
            exit_code
        """
        dest = tmpl_value.get(KDEST, "")
        bnbr = tmpl_value.get("backup", 0)
        bpath = tmpl_value.get("backup_dir")
        template = Template(self._read_template_source(tmpl_value.get("src", "")))
        tfile = NamedTemporaryFile(
            mode="w",
            encoding="utf-8",
            suffix=None,
            delete=False,
        )
        tfile.write(template.render(template_dict=tmpl_var))
        tfile.close()
        return self._write_output(Path(dest), Path(tfile.name), bnbr, bpath)

    def _write_output(
        self,
        dpath: Path,
        tpath: Path,
        bnbr: int,
        bpath: Optional[Path] = None,
    ) -> bool:
        """Write output to output file, unlink temporary file.

        Parameters
        ----------
        dpath : Path
            Path to the output file
        tpath : Path
            Path to the temporary generated template file
        bnbr : int
            Number of backups to keep if any
        bpath : Optional[Path]
            Directory to store the backups in

        Returns
        -------
        bool : True if target file replaced
        """
        retval = False
        if dpath.exists():
            diff = filecmp.cmp(dpath, tpath)
            exists = True
        else:
            exists = False
            diff = False
        self.debug(
            "dest: {0} exists: {1} diff: {2}".format(str(dpath), exists, diff),
        )
        if not diff:
            if not self.options.get("noop", False):
                if exists:
                    self._backup_file(str(dpath), bnbr, bpath)
                copy2(tpath, dpath)
                self.info("Updated file: {0}".format(str(dpath)))
                retval = True
        if not self.isdebug():
            unlink_path(tpath, missing_ok=True)
        return retval

    def _backup_file(self, dest: str, bnbr: int, bpath: Optional[Path] = None) -> None:
        """Backup the output before replacing.

        Parameters
        ----------
        dest : str
            Path to output file.
        bnbr : int
            Number of backup files to keep (if any)
        bpath : Optional[Path]
            Directory to store the backups in
        """
        if bnbr:
            if bpath is None:
                version_file(dest, "rename", bnbr, self.isdebug())
            else:
                version_file(dest, "rename", bnbr, self.isdebug(), str(bpath))

    def _read_template_source(self, source: str) -> str:
        """Retruns the template data from the given source.

        Parameters
        ----------
        source : str
            Path to the template source

        Returns
        -------
        str
            The template data
        """
        with open(source, "r") as jinja_file:
            return jinja_file.read()

    def _verify_template_required_keys(
        self,
        tmpl_name: str,
        tmpl_value: StrAnyDict,
    ) -> bool:
        """Verify that the template required keys exist."""
        for key in ("src", KDEST):
            kv = tmpl_value.get(key)
            if kv is None:
                self.error(
                    "Template {0} does not have a {1} key!!".format(tmpl_name, key),
                )
                return False
        return True

    def _verify_target(self, dest: str) -> bool:
        """Verifies the output file.

        Parameters
        ----------
        dest : str
            Path to the output file

        Returns
        -------
        bool
            True if target directory exists and writable and file does not exist
            True if target exists, is a file, and writable
        """
        dspec = Path(dest)
        if dspec.exists():
            if not dspec.is_file():
                self.error("Template dest '{0}' not a file!!".format(dest))
                return False
            if not access(dspec, W_OK):
                self.error("Template dest '{0}' not readable!!".format(dest))
                return False
        else:
            return self._verify_target_directory(dspec.parent)
        return True

    def _verify_target_directory(self, pspec: Path) -> bool:
        """Verifies the target directory.

        Parameters
        ----------
        pspec : Path
            Path to the target directory

        Returns
        -------
        bool
            True if exists, is directory and is writable.
        """
        retval, error_message = verify_directory(pspec)
        if not retval:
            self.error(error_message)
        return retval

    def _verify_template_source(self, source: str) -> bool:
        """Verifies the template source.

        Parameters
        ----------
        source : str
            Path to the template source

        Returns
        -------
        bool
            True if source exists, is a file, is readable
        """
        spec = Path(source)
        if not spec.exists():
            self.error("Template src '{0}' not found!!".format(source))
            return False
        if not spec.is_file():
            self.error("Template src '{0}' not a file!!".format(source))
            return False
        if not access(spec, R_OK):
            self.error("Template src '{0}' not readable!!".format(source))
            return False
        return True



"""Top-level module for wtforglib Library."""
import platform
from inspect import stack
from typing import Tuple, Union

from wtforglib.functions import WINDOZE, windoze_not_implemented

if platform.system() != WINDOZE:
    from grp import getgrall, getgrgid  # noqa: WPS433
    from pwd import getpwnam, getpwuid  # noqa: WPS433


def get_user_name(uid: int) -> str:
    """Returns the user name of the user identified by uid.

    Parameters
    ----------
    uid : int
        User id

    Returns
    -------
    str
        User name
    """
    windoze_not_implemented(stack()[0][3])
    return getpwuid(uid)[0]


def get_user_groups(user: Union[str, int]) -> Tuple[str, ...]:
    """Returns a tuple of groups user belongs to.

    Parameters
    ----------
    user : Union[str, int]
        User name or uid

    Returns
    -------
    Tuple[str,...]
        Groups user belongs to
    """
    windoze_not_implemented(stack()[0][3])
    if isinstance(user, int):
        user_nm = get_user_name(int(user))
    else:
        user_nm = user
    groups = [gg.gr_name for gg in getgrall() if user_nm in gg.gr_mem]
    gid = getpwnam(user_nm).pw_gid
    groups.append(getgrgid(gid).gr_name)
    return tuple(groups)



"""Top-level module kinds for wtforglib.

The kinds module supports defintions of types used in
the wtforglib package and used in various projects
the author provides.

Types
-----
KeyType
    Genery mapping key type
ValueType
    Generic mapping value type
Fspec
    Fspec = Union[str, PathLike[str]]
StrAnyDict
    StrAnyDict = Dict[str, Any]
"""

import sys
from os import PathLike
from typing import Any, Dict, TypeVar, Union

KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")


if sys.version_info >= (3, 9):
    Fspec = Union[str, PathLike[str]]
else:
    Fspec = Union[str, "PathLike[str]"]

StrAnyDict = Dict[str, Any]  # type: ignore
StrStrDict = Dict[str, str]
StrStrInt = Dict[str, Union[int, str]]
StrStrIntBool = Dict[str, Union[int, str, bool]]



"""Top-level module for wtforglib Library."""

# vim:ft=py noqa: E800
import platform

from wtforglib.functions import WINDOZE

if platform.system() == WINDOZE:
    import ctypes  # noqa: WPS433

    SUPER = "Administrator"
else:
    from os import geteuid  # noqa: WPS433

    SUPER = "super user"


class AdminStateUnknownError(Exception):
    """Cannot determine whether the user is an admin."""


def issuper() -> bool:
    """Return True/False depending on process euid.

    Returns
    -------
        bool : true if process euid is 0
    """
    try:
        is_super = geteuid() == 0
    except AttributeError:
        is_super = (
            ctypes.windll.shell32.IsUserAnAdmin() != 0  # type: ignore[attr-defined]
        )
    return is_super


def requires_super_user(prefix: str = "Specified action") -> None:
    """Raises exception if process isn't euid 0.

    Parameters
    ----------
    prefix : str
        Text to prefix to exception message

    Raises
    ------
    PermissionError
        if process euid is not 0
    """
    if not issuper():  # pragma no cover
        raise PermissionError("{0} requires {1} priviledges!".format(prefix, SUPER))



"""Top-level module for wtforglib Library."""
import platform
import re
import shutil
from inspect import stack
from pathlib import Path
from typing import Optional, Union

from wtforglib.functions import WINDOZE, windoze_not_implemented
from wtforglib.kinds import Fspec

OwnGrpId = Optional[Union[str, int]]


if platform.system() != WINDOZE:
    from grp import getgrgid  # noqa: WPS433
    from pwd import getpwuid  # noqa: WPS433


def set_file_perms(tgt: Fspec, mode: str) -> bool:  # noqa: WPS231
    """Sets the owner group permissions for a posix file if needed.

    Parameters
    ----------
    tgt : Fspec
        Path to the target file
    mode : str
        Mode of the target file

    Returns
    -------
    bool : True if changes are made to the target

    Raises
    ------
    ValueError
        When mode is not valid
    """
    if mode:
        target = Path(tgt)
        mode_length = len(mode)
        if mode_length > 4 or mode_length < 3:
            raise ValueError("Invalid mode string length. Must be 3 or 4 characters.")
        if not re.match("[0-7]+$", mode):
            raise ValueError("Invalid mode string. Must be 3 or 4 octal digits.")
        windoze_not_implemented(stack()[0][3])
        # convert mode to octal string right 4 chars
        cmode = str(oct(target.stat().st_mode))[-mode_length:]
        if cmode != mode:
            omode = int(mode, 8)
            target.chmod(omode)
            return True
    return False


def get_new_owner(tgt: Fspec, own: OwnGrpId) -> Optional[str]:
    """Sets the owner for a posix file in needed.

    Parameters
    ----------
    tgt : Fspec
        Path to the target file
    own : OwnGrpId
        Owner of the target file

    Returns
    -------
    Optional[str] : new_owner or None
    """
    windoze_not_implemented(stack()[0][3])
    new_owner: Optional[str]
    if own:
        if isinstance(own, int):
            new_owner = getpwuid(own)[0]
        else:
            new_owner = str(own)
        if Path(tgt).owner() == new_owner:
            new_owner = None
    else:
        new_owner = None
    return new_owner


def get_new_group(tgt: Fspec, grp: OwnGrpId) -> Optional[str]:
    """Sets the owner for a posix file in needed.

    Parameters
    ----------
    tgt : Fspec
        Path to the target file
    grp : OwnGrpId
        Group of the target file

    Returns
    -------
    Optional[str] : new_group or None
    """
    windoze_not_implemented(stack()[0][3])
    new_group: Optional[str]
    if grp:
        if isinstance(grp, int):
            new_group = getgrgid(grp)[0]
        else:
            new_group = str(grp)
        if Path(tgt).group() == new_group:
            new_group = None
    else:
        new_group = None
    return new_group


def set_owner_group(tgt: Fspec, own: OwnGrpId, grp: OwnGrpId) -> bool:
    """Sets the owner group for a posix file in needed.

    Parameters
    ----------
    tgt : Fspec
        Path to the target file
    own : str
        Owner of the target file
    grp : str
        Group of the target file

    Returns
    -------
    bool : True if changes are made to the target
    """
    windoze_not_implemented(stack()[0][3])
    target = Path(tgt)
    new_owner = get_new_owner(target, own)
    new_group = get_new_group(target, grp)
    if (new_group is not None) or (new_owner is not None):
        shutil.chown(target, new_owner, new_group)  # type: ignore
        return True
    return False


def set_owner_group_perms(tgt: Fspec, own: OwnGrpId, grp: OwnGrpId, mode: str) -> bool:
    """Sets the owner group permissions for a posix file if needed.

    Parameters
    ----------
    tgt : Fspec
        Path to the target file
    own : str
        Owner of the target file
    grp : str
        Group of the target file
    mode : str
        Mode of the target file

    Returns
    -------
    bool : True if changes are made to the target
    """
    windoze_not_implemented(stack()[0][3])
    og_changed = set_owner_group(tgt, own, grp)
    fp_changed = set_file_perms(tgt, mode)
    return og_changed or fp_changed



from urllib.parse import urlparse


def url_validator(url: str) -> bool:
    """Validate url."""
    try:
        parsed = urlparse(url)
        return all([parsed.scheme, parsed.netloc])
    except Exception:
        return False



"""
Top-level module scribe for wtforglib.

Classes:

    Scribe

Functions:

    None

Misc variables:

    LOG_LEVELS
"""


import logging
import platform
from logging.handlers import SysLogHandler
from types import MappingProxyType
from typing import List

LOG_LEVELS = MappingProxyType(
    {
        "crit": logging.CRITICAL,
        "error": logging.ERROR,
        "warn": logging.WARN,
        "info": logging.INFO,
        "debug": logging.DEBUG,
    },
)


def log_level_name(level: int, default: str = "info") -> str:
    """Return logger level name.

    Parameters
    ----------
    level : int
        Level number
    default : str, optional
        Default if not matched, by default "info"

    Returns
    -------
    str
        Level name
    """
    for key, valor in LOG_LEVELS.items():
        if level == valor:
            return key
    return default


class Scribe(object):
    """
    A class to create a logging system interface.

    ...

    Attributes
    ----------
    logger : Any
    """

    def __init__(self, **kwargs) -> None:
        r"""Initialize a logging system interface.

        :param \**kwargs:
            See below

        :Keyword Arguments:
            * *level* (``str``) --
                one of: 'debug', 'info', 'warn', 'error', 'critical', default: 'info'
            * *logfn* (``str``) --
                file name of log file, default ""
            * *sylognm* (``str``) --
                annotation used in syslog to identfiy records, default ""
            * *screen* (``bool``) --
                flag specifyig whether to log to screen, default False
            * *name* (``str``) --
                unique identifier for this loggin instance, default "wtfscribe"
        """
        warnings: List[str] = []
        self.name = kwargs.get("name", "wtfscribe")
        level = kwargs.get("level", "info")
        self.logfn = kwargs.get("logfn", "")
        self.syslognm = kwargs.get("syslognm", "")
        screen = kwargs.get("screen", False)
        self.logger = logging.getLogger(self.name)

        if level not in LOG_LEVELS:
            self.log_level = logging.INFO
            warnings.append("Invalid log level '{0}' changed to 'info'".format(level))
        else:
            self.log_level = LOG_LEVELS[level]
        self.logger.setLevel(self.log_level)
        self.handlers_nbr = 0
        self._add_file_handler()
        if platform.system() != "Windows":
            self._add_syslog_handler()
        self._add_screen_handler(screen)
        for warning in warnings:
            self.logger.warning(warning)

    def _add_file_handler(self) -> bool:
        """Initialize a file log handler.

        Returns
        -------
        bool
        """
        if self.logfn:
            filelog = logging.FileHandler(filename=self.logfn)
            filelog.setLevel(self.log_level)
            formatter = logging.Formatter(
                "%(asctime)s %(levelname)s: %(message)s",
                "%Y-%m-%d %H:%M:%S",
            )
            filelog.setFormatter(formatter)
            self.logger.addHandler(filelog)
            self.handlers_nbr += 1
            return True
        return False

    def _add_syslog_handler(self) -> bool:
        """Initialize a syslog handler.

        Returns
        -------
        bool
        """
        if self.syslognm:
            syslog = SysLogHandler(address="/dev/log")
            syslog.setFormatter(
                logging.Formatter(
                    "[{0}] %(levelname)8s: %(message)s".format(self.syslognm),
                ),
            )
            # don't send debug to syslog
            syslog_level = (
                logging.INFO if self.log_level < logging.INFO else self.log_level
            )
            syslog.setLevel(syslog_level)
            self.logger.addHandler(syslog)
            self.handlers_nbr += 1
            return True
        return False

    def _add_screen_handler(self, screen: bool) -> bool:
        """
        Add a screen log handler.

        Parameters
        ----------
        screen : bool
            specify whether to log to screen

        Returns
        -------
        bool

        """
        if not self.handlers_nbr or screen:
            console = logging.StreamHandler()
            # set a format which is simpler for console use
            # tell the handler to use this format
            console.setFormatter(logging.Formatter("%(levelname)8s: %(message)s"))
            console.setLevel(self.log_level)
            self.logger.addHandler(console)
            self.handlers_nbr += 1
            return True
        return False
